#include "fancygraph.h"
#include "qevent.h"

#include <QTimer>
#include <QPainter>
#include <QPainterPath>
#include <QDateTime>
#include <cmath>

FancyGraph::FancyGraph(QWidget *parent)
    : QWidget{parent}
{
    timer = new QTimer(this);
    connect(timer, &QTimer::timeout, this, QOverload<>::of(&FancyGraph::repaint));
    timer->start(1000/60);
}

QPointF FancyGraph::dataToScreen(QPointF dataPoint, QRectF paintZone)
{
    double time = QDateTime::currentMSecsSinceEpoch() / 1000.0;

    QPointF screenPoint;

    screenPoint.setX(paintZone.left() + (1-((time - dataPoint.x())/historyLength)) * paintZone.width());
    screenPoint.setY(paintZone.top() + (1-dataPoint.y()) * paintZone.height());

    return screenPoint;
}

void FancyGraph::paintEvent(QPaintEvent* event)
{
    QRectF paintZone = rect().adjusted(6,6,-6,-6);
    QColor colour = QColor(217,0,74);

    QPainter painter(this);
    painter.setRenderHint(QPainter::Antialiasing);
//    painter.setRenderHint(QPainter::HighQualityAntialiasing);
    painter.setClipRegion(event->region());
//    painter.begin(this);/

    double time = QDateTime::currentMSecsSinceEpoch() / 1000.0;

    // trim data points from too far in the past
    while (!points.empty() && points.first().x() < time - (historyLength * 1.1))
    {
        points.removeFirst();
    }

    if (!points.empty())
    {
        QPainterPath path;
        path.moveTo(dataToScreen(points.first(), paintZone));
        for (QPointF point: points)
        {
            path.lineTo(dataToScreen(point, paintZone));
        }

//        path.addEllipse(end, 3, 3);

        painter.setPen(QPen(QBrush(colour), 2.5, Qt::PenStyle::SolidLine, Qt::PenCapStyle::RoundCap, Qt::RoundJoin));
        painter.drawPath(path);

        QPointF end = dataToScreen(points.last(), paintZone);
        end.setX(paintZone.right());
        painter.setBrush(QBrush(colour));
        painter.drawEllipse(end, 3, 3);
    }

//    painter.end();
}

void FancyGraph::addPoint(double value)
{
    if (!std::isnan(value))
    {
        QPointF point(QDateTime::currentMSecsSinceEpoch() / 1000.0, value);
        points.append(point);
    }
}
