#ifndef GUI_H
#define GUI_H

#include <QMainWindow>
#include <QFile>
#include <QTextStream>
#include <QTimer>
#include <QMutex>
#include <QThread>
#include <QCloseEvent>
#include <QList>
#include <QMetaMethod>

#include "../rulemodule.h"
#include "telemetry.h"

#include "../libraries/buttplugCpp/include/buttplugclient.h"

const QString CANT_CONNECT_MESSAGE = "<html><head/><body><p>Can't connect to Intiface server!</p><p>Make sure that <a href=\"https://intiface.com/central/\"><span style=\" text-decoration: underline; color:#8ab4f8;\">Intiface Central</span></a> is installed and running.</p><p><span style=\" font-style:italic;\">[ If Intiface is on a different device, enter the address (ws://...) on the 'Config' tab. ]</span></p></body></html>";

namespace Ui {
class Gui;
}

class Gui : public QMainWindow
{
    Q_OBJECT

static const int UPDATE_INTERVAL_MS = 100;

public:
    explicit Gui(QWidget *parent = nullptr);
    ~Gui();

    void loadSettings();
    void saveSettings();
    void shutdown();
    void logLine(QString line);

    void receiveTelemetry(TelemetryState tele);
    void gameplayEvent(QString event, QMap<QString, QVariant> attributes);
    void setPause(bool pause);

    virtual void closeEvent(QCloseEvent* event) override;

protected slots:
    void update();
    void updateStatus();
    void rulesStatusUpdate(QString status);

signals:
    void shuttingDown();

private slots:
    void on_jobCompleteTestButton_pressed();

private:
    Ui::Gui* ui;

    Client* client;
    QFile* file;
    QTextStream* stream;
    QTimer* timer;

    QMutex telemetryMutex;
    TelemetryState latestTelemetry;
    double distanceCounter = 0.0;
    bool paused = true;

    scs_timestamp_t lastTime = 0;

    QList<RuleModule*> modules;
    SpeedModule speedMod;
    RPMModule   rpmMod;
    CrimeModule crimeMod;
    JobModule   jobMod;
    DistanceModule distMod;

    QString rulesStatusText;

    void callbackFunction(const mhl::Messages msg);
    std::function<void(const mhl::Messages)> callbackFunctionPtr = [this](const mhl::Messages msg){callbackFunction(msg);};

//    static void doubleConnect(QObject* objectA, QObject* objectB, const QMetaMethod aSignal, const QMetaMethod aSlot, const QMetaMethod bSignal, const QMetaMethod bSlot );



};

#endif // GUI_H
