// Windows stuff.
#ifdef _WIN32
#  define WINVER 0x0500
#  define _WIN32_WINNT 0x0500
#  include <windows.h>
#endif

// SDK
#include "../libraries/scs-sdk/include/scssdk_telemetry.h"
#include "../libraries/scs-sdk/include/eurotrucks2/scssdk_eut2.h"
#include "../libraries/scs-sdk/include/eurotrucks2/scssdk_telemetry_eut2.h"
#include "../libraries/scs-sdk/include/amtrucks/scssdk_ats.h"
#include "../libraries/scs-sdk/include/amtrucks/scssdk_telemetry_ats.h"

#include "telemetry.h"
#include "qapphost.h"
//#include "eurofucksimulator.h"
#include "gui/gui.h"

#include <QVector3D>
#include <QMatrix4x4>
#include <QQuaternion>

//#include <QApplication>
//#include <pthread.h>

#define UNUSED(x)

scs_timestamp_t last_timestamp = static_cast<scs_timestamp_t>(-1);
TelemetryState telemetry;

QAppHost* host = nullptr;
//EuroFuckSimulator* eu = nullptr;
Gui* gui = nullptr;

QVariant value_to_variant(scs_value_t value)
{
    QVariant ret;
    switch (value.type)
    {
    case SCS_VALUE_TYPE_INVALID:
    {
        ret = QVariant();
        break;
    }
    case SCS_VALUE_TYPE_bool: {
        ret = QVariant::fromValue(value.value_bool.value);
        break;
    }
    case SCS_VALUE_TYPE_s32: {
        ret = QVariant::fromValue(static_cast<int>(value.value_s32.value));
        break;
    }
    case SCS_VALUE_TYPE_u32: {
        ret = QVariant::fromValue(static_cast<unsigned>(value.value_u32.value));
        break;
    }
    case SCS_VALUE_TYPE_s64: {
        ret = QVariant::fromValue(value.value_s64.value);
        break;
    }
    case SCS_VALUE_TYPE_u64: {
        ret = QVariant::fromValue(value.value_u64.value);
        break;
    }
    case SCS_VALUE_TYPE_float: {
        ret = QVariant::fromValue(value.value_float.value);
        break;
    }
    case SCS_VALUE_TYPE_double: {
        ret = QVariant::fromValue(value.value_double.value);
        break;
    }
    case SCS_VALUE_TYPE_fvector: {
        ret = QVariant::fromValue(QVector3D(value.value_fvector.x,value.value_fvector.y,value.value_fvector.z));
        break;
    }
    case SCS_VALUE_TYPE_dvector: {
        ret = QVariant::fromValue(QVector3D(value.value_dvector.x,value.value_dvector.y,value.value_dvector.z));
        break;
    }
    case SCS_VALUE_TYPE_euler: {
        ret = QVariant::fromValue(QQuaternion::fromEulerAngles(value.value_euler.heading * 360.0f,value.value_euler.pitch * 360.0f,value.value_euler.roll * 360.0f));
        break;
    }
//    case SCS_VALUE_TYPE_fplacement: {
////        QMatrix4x4::QMatrix4x4
////        log_line(
////            "fplacement = (%f,%f,%f) h:%f p:%f r:%f",
////            value.value_fplacement.position.x,
////            value.value_fplacement.position.y,
////            value.value_fplacement.position.z,
////            value.value_fplacement.orientation.heading * 360.0f,
////            value.value_fplacement.orientation.pitch * 360.0f,
////            value.value_fplacement.orientation.roll * 360.0f
////            );
//        break;
//    }
//    case SCS_VALUE_TYPE_dplacement: {
////        log_line(
////            "dplacement = (%f,%f,%f) h:%f p:%f r:%f",
////            value.value_dplacement.position.x,
////            value.value_dplacement.position.y,
////            value.value_dplacement.position.z,
////            value.value_dplacement.orientation.heading * 360.0f,
////            value.value_dplacement.orientation.pitch * 360.0f,
////            value.value_dplacement.orientation.roll * 360.0f
////            );
//        break;
//    }
    case SCS_VALUE_TYPE_string: {
        ret = QVariant::fromValue(QString(value.value_string.value));
        break;
    }
    default: {
        ret = QVariant();
        break;
    }
    }
    return ret;
}

QMap<QString, QVariant> attributes_to_map(const scs_named_value_t *const attributes)
{
    QMap<QString, QVariant> ret;
    for (const scs_named_value_t *current = attributes; current->name; ++current)
    {
        QString name = QString::fromStdString(current->name);
        QVariant value = value_to_variant(current->value);
        ret[name] = value;
    }
    return ret;
}

SCSAPI_VOID telemetry_frame_start(const scs_event_t UNUSED(event), const void *const event_info, const scs_context_t UNUSED(context))
{
    const struct scs_telemetry_frame_start_t *const info = static_cast<const scs_telemetry_frame_start_t *>(event_info);

    int delta = info->paused_simulation_time - last_timestamp;
    telemetry.timestamp += (delta);
//    telemetry.time_delta = delta;
    last_timestamp = info->paused_simulation_time;
}

SCSAPI_VOID telemetry_frame_end(const scs_event_t UNUSED(event), const void *const UNUSED(event_info), const scs_context_t UNUSED(context))
{
    if (gui != nullptr)
        gui->receiveTelemetry(telemetry);
}

SCSAPI_VOID telemetry_pause(const scs_event_t event, const void *const UNUSED(event_info), const scs_context_t UNUSED(context))
{
    bool output_paused = (event == SCS_TELEMETRY_EVENT_paused);
    if (gui != nullptr)
        gui->setPause(output_paused);
}

SCSAPI_VOID telemetry_configuration(const scs_event_t event, const void *const event_info, const scs_context_t UNUSED(context))
{
    const struct scs_telemetry_configuration_t *const info = static_cast<const scs_telemetry_configuration_t *>(event_info);
//    qDebug() << attributes_to_map(info->attributes);
}

SCSAPI_VOID telemetry_gameplay_event(const scs_event_t event, const void *const event_info, const scs_context_t UNUSED(context))
{
    const struct scs_telemetry_gameplay_event_t *const info = static_cast<const scs_telemetry_gameplay_event_t *>(event_info);

    QMap<QString,QVariant> map = attributes_to_map(info->attributes);

    gui->gameplayEvent(QString::fromStdString(info->id), map);
}

SCSAPI_VOID telemetry_store_orientation(const scs_string_t name, const scs_u32_t index, const scs_value_t *const value, const scs_context_t context)
{
    assert(context);
    TelemetryState *const state = static_cast<TelemetryState *>(context);

    // This callback was registered with the SCS_TELEMETRY_CHANNEL_FLAG_no_value flag
    // so it is called even when the value is not available.

    if (! value) {
        state->orientation_available = false;
        return;
    }

    assert(value);
    assert(value->type == SCS_VALUE_TYPE_euler);
    state->orientation_available = true;
    state->heading = value->value_euler.heading * 360.0f;
    state->pitch = value->value_euler.pitch * 360.0f;
    state->roll = value->value_euler.roll * 360.0f;
}

SCSAPI_VOID telemetry_store_float(const scs_string_t name, const scs_u32_t index, const scs_value_t *const value, const scs_context_t context)
{
    // The SCS_TELEMETRY_CHANNEL_FLAG_no_value flag was not provided during registration
    // so this callback is only called when a valid value is available.

    assert(value);
    assert(value->type == SCS_VALUE_TYPE_float);
    assert(context);
    *static_cast<float *>(context) = value->value_float.value;
}

SCSAPI_VOID telemetry_store_s32(const scs_string_t name, const scs_u32_t index, const scs_value_t *const value, const scs_context_t context)
{
    // The SCS_TELEMETRY_CHANNEL_FLAG_no_value flag was not provided during registration
    // so this callback is only called when a valid value is available.

    assert(value);
    assert(value->type == SCS_VALUE_TYPE_s32);
    assert(context);
    *static_cast<int *>(context) = value->value_s32.value;
}

void boot()
{
    if (gui == nullptr)
    {
        gui = new Gui();
        gui->setVisible(true);
    }
}

SCSAPI_RESULT scs_telemetry_init(const scs_u32_t version, const scs_telemetry_init_params_t *const params)
{
    if (version != SCS_TELEMETRY_VERSION_1_01) {
        return SCS_RESULT_unsupported;
    }

    try
    {
        if (host == nullptr)
        {
            host = new QAppHost(&boot);
        }
    }
    catch (std::exception ex)
    {
        std::cerr << "Killed by exception: " << ex.what() << std::endl;
        return SCS_RESULT_generic_error;
    }

    const scs_telemetry_init_params_v101_t *const version_params = static_cast<const scs_telemetry_init_params_v101_t *>(params);

    const bool events_registered =
        (version_params->register_for_event(SCS_TELEMETRY_EVENT_frame_start, telemetry_frame_start, nullptr) == SCS_RESULT_ok) &&
        (version_params->register_for_event(SCS_TELEMETRY_EVENT_frame_end, telemetry_frame_end, nullptr) == SCS_RESULT_ok) &&
        (version_params->register_for_event(SCS_TELEMETRY_EVENT_paused, telemetry_pause, nullptr) == SCS_RESULT_ok) &&
                                   (version_params->register_for_event(SCS_TELEMETRY_EVENT_started, telemetry_pause, nullptr) == SCS_RESULT_ok);

    if (! events_registered)
    {
        version_params->common.log(SCS_LOG_TYPE_error, "Unable to register event callbacks");
        return SCS_RESULT_generic_error;
    }

    version_params->register_for_event(SCS_TELEMETRY_EVENT_configuration, telemetry_configuration, nullptr);
    version_params->register_for_event(SCS_TELEMETRY_EVENT_gameplay, telemetry_gameplay_event, nullptr);

    version_params->register_for_channel(SCS_TELEMETRY_TRUCK_CHANNEL_speed, SCS_U32_NIL, SCS_VALUE_TYPE_float, SCS_TELEMETRY_CHANNEL_FLAG_none, telemetry_store_float, &telemetry.speed);
    version_params->register_for_channel(SCS_TELEMETRY_TRUCK_CHANNEL_engine_rpm, SCS_U32_NIL, SCS_VALUE_TYPE_float, SCS_TELEMETRY_CHANNEL_FLAG_none, telemetry_store_float, &telemetry.rpm);
    version_params->register_for_channel(SCS_TELEMETRY_TRUCK_CHANNEL_engine_gear, SCS_U32_NIL, SCS_VALUE_TYPE_s32, SCS_TELEMETRY_CHANNEL_FLAG_none, telemetry_store_s32, &telemetry.gear);



    return SCS_RESULT_ok;
}

SCSAPI_VOID scs_telemetry_shutdown(void)
{
    if (gui != nullptr)
    {
        gui->setVisible(false);
        gui->saveSettings();
        gui->shutdown();
        delete gui;
    }
    if (host != nullptr)
    {
        host->shutdown();
        delete host;
    }
}
